from collections import defaultdict
import logging
import os
import time

import numpy as np

try:
    from lal.gpstime import tconvert
except ImportError:
    tconvert = None

from . import configparser
from . import plots
from . import names
from . import features
from . import factories
from . import utils
from .series import combine_series, extract_segments_from_series

from ligo.scald import report
from ligo.segments import segment, segmentlist

__process_name__ = "iDQ"

logger = logging.getLogger("idq")

DEFAULT_PGLITCH_THRESHOLD = 0.9
DEFAULT_FIGTYPE = "png"
DEFAULT_LINKS = [
    ("https://git.ligo.org/lscsoft/iDQ", "iDQ source code"),
    ("https://git.ligo.org/lscsoft/iDQ/wikis", "iDQ wiki"),
    ("https://docs.ligo.org/lscsoft/iDQ/", "iDQ docs"),
]


class Report(object):
    """Generate everything needed for a report."""

    def __init__(
        self,
        config_path,
        start,
        end,
        segments=None,
        t0=None,
        zoom_start=None,
        zoom_end=None,
        skip_timeseries=False,
    ):
        self._start = start
        self._end = end
        self._t0 = t0
        self._zoom_start = zoom_start
        self._zoom_end = zoom_end

        if segments:
            self._segments = segments
        else:
            self._segments = segmentlist([segment(start, end)])

        # load configuration
        self._config_path = os.path.abspath(os.path.realpath(config_path))
        self._config = configparser.path2config(config_path)
        self._instrument = self.config.instrument
        self._tag = self.config.tag

        self.figtype = DEFAULT_FIGTYPE
        self.pglitch_thr = DEFAULT_PGLITCH_THRESHOLD
        self.skip_timeseries = skip_timeseries

        self.legend = self._config.report.get("legend", True)
        self.annotate_gch = self._config.report.get("annotate_gch", False)
        self.annotate_auc = self._config.report.get("annotate_auc", False)
        self.overview_only = self._config.report.get("overview_only", False)
        self.single_calib_plots = self._config.report.get("single_calib_plots", True)

        # check frequency bounds
        if "frequency" in self.config.features:
            target_bounds = configparser.config2bounds(
                self.config.samples["target_bounds"]
            )
            frequency = self.config.features["frequency"]
            assert (
                frequency in target_bounds
            ), f"must specify a frequency range ({frequency}) within target_bounds"
            freq_min, freq_max = target_bounds[frequency]
            assert isinstance(freq_min, int) and isinstance(
                freq_max, int
            ), "frequency bounds must be integers!"
            self._freq_bounds = freq_min, freq_max
        else:
            self._freq_bounds = None

    @property
    def config(self):
        return self._config

    @property
    def config_path(self):
        return self._config_path

    @property
    def instrument(self):
        return self._instrument

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def segs(self):
        return self._segments

    @property
    def zoom_start(self):
        return self._zoom_start if (self._zoom_start is not None) else self._start

    @property
    def zoom_end(self):
        return self._zoom_end if (self._zoom_end is not None) else self._end

    @property
    def t0(self):
        return (
            self._t0
            if (self._t0 is not None)
            else 0.5 * (self.zoom_start + self.zoom_end)
        )

    @property
    def tag(self):
        return self._tag

    def create_runtime_overview(self):
        """Generate runtime statistics and return a table element for the report."""
        start_date = tconvert(self.start)
        end_date = tconvert(self.end)
        livetime = utils.livetime(self.segs)
        columns = ["", ""]
        rows = [
            ["Start Time", f"{self.start:.3f} ({start_date})"],
            ["End Time", f"{self.end:.3f} ({end_date})"],
            ["Livetime (sec)", f"{livetime:.1f}"],
            [
                "Detector Duty Cycle",
                f"{(100 * livetime) / (self.end - self.start):.1f} %",
            ],
        ]
        return report.FlatTable("Summary", columns=columns, rows=rows)

    def create_classifier_overview(
        self,
        nicknames,
        datasets,
        models,
        calibmaps,
        series=None,
    ):
        """Generate classifier overview statistics and return a table element
        for the report."""
        columns = [
            "nickname",
            "No. glitch samples",
            "No. clean samples",
            "No. models",
            "No. calibration maps",
        ]
        if series is not None:
            columns += [
                f"Time with p(glitch) >= {self.pglitch_thr:.3f}",
                "Walltime Duty Cycle",
                "Science Duty Cycle",
            ]

        rows = []
        for nickname in nicknames:
            # do some quick data manipulations
            gch, cln = datasets[nickname].vectors2classes()
            row = [
                nickname,
                "%d" % len(gch),
                "%d" % len(cln),
                "%d" % len(models[nickname]),
                "%d" % len(calibmaps[nickname]),
            ]

            if series is not None:
                times = np.concatenate([s.times for s in series[nickname]])
                pglitch = np.concatenate([s.pglitch.series for s in series[nickname]])
                pglitch = pglitch[(self.start <= times) * (times < self.end)]
                pglitch_frac = np.sum(pglitch >= self.pglitch_thr) / len(pglitch)

                row.append(f"{100. * pglitch_frac:.3f} %")
                these_segs = extract_segments_from_series(series[nickname])
                # series might slightly over-cover requested segments
                walltime_duty_cycle = 100 * min(
                    1.0, utils.livetime(these_segs) / (self.end - self.start)
                )
                science_duty_cycle = (
                    100
                    * utils.livetime(these_segs & self.segs)
                    / utils.livetime(self.segs)
                )
                row.append(f"{walltime_duty_cycle:.2f} %")
                row.append(f"{science_duty_cycle:.2f} %")

            rows.append(row)

        return report.FlatTable("Basic Statistics", columns=columns, rows=rows)

    def report(self, nicknames, reportdir):
        """Generate a full report."""
        # act like a DiskReporter...
        reportdir = names.start_end2fragmented_dir(
            self.start, self.end, rootdir=reportdir
        )
        os.makedirs(reportdir, exist_ok=True)

        logger.info(
            "generating report for %s within [%.3f, %.3f) with %s"
            % (", ".join(nicknames), self.start, self.end, self.config_path)
        )

        # find evaluated datasets
        logger.info(
            "finding evaluated datasets within [%.3f, %.3f)" % (self.start, self.end)
        )
        datasets = {
            nickname: find_datasets(
                self.config, self.start, self.end, nickname, segments=self.segs
            )
            for nickname in nicknames
        }
        for nickname in nicknames:
            gch, cln = datasets[nickname].vectors2classes()
            logger.info(
                "found %d gch samples and %d cln samples for %s"
                % (len(gch), len(cln), nickname)
            )

        # find timeseries, models and calibration maps
        models = {}
        calibmaps = {}

        # just grab everything available
        if self.skip_timeseries:
            series = None

            logger.info("finding models within [%.3f, %.3f)" % (self.start, self.end))
            for nickname in nicknames:
                models[nickname] = find_models(
                    self.config, self.start, self.end, nickname
                )
                logger.info(
                    "found %d models for %s" % (len(models[nickname]), nickname)
                )

            logger.info(
                "finding calibration maps within [%.3f, %.3f)" % (self.start, self.end)
            )
            for nickname in nicknames:
                calibmaps[nickname] = find_calibration_maps(
                    self.config, self.start, self.end, nickname
                )
                logger.info(
                    "found %d calibration maps for %s"
                    % (len(calibmaps[nickname]), nickname)
                )

        # grab series and only load the ones used to generate this data
        else:
            logger.info(
                "finding timeseries within [%.3f, %.3f)" % (self.start, self.end)
            )
            series = {}
            for nickname in nicknames:
                series[nickname] = find_timeseries(
                    self.config, self.start, self.end, nickname
                )
                logger.info(
                    "found %d timeseries for %s" % (len(series[nickname]), nickname)
                )

            logger.info(
                "finding models within [%.3f, %.3f) based on timeseries"
                % (self.start, self.end)
            )
            for nickname in nicknames:
                model_ids = {s.model_id for s in series[nickname]}
                models[nickname] = find_models(
                    self.config, self.start, self.end, nickname, hashes=model_ids
                )
                logger.info(
                    "found %d models out of %d requested for %s"
                    % (len(models[nickname]), len(model_ids), nickname)
                )

            logger.info(
                "finding calibration maps within [%.3f, %.3f) based on timeseries"
                % (self.start, self.end)
            )
            for nickname in nicknames:
                calibration_ids = {s.calibration_id for s in series[nickname]}
                calibmaps[nickname] = find_calibration_maps(
                    self.config, self.start, self.end, nickname, hashes=calibration_ids
                )
                logger.info(
                    "found %d models out of %d requested for %s"
                    % (len(calibmaps[nickname]), len(calibration_ids), nickname)
                )

        # instantiate the report
        report_name = "%s Report at %s (%s)" % (
            __process_name__,
            self.instrument,
            self.tag,
        )
        thisreport = report.Report(
            report_name, theme="gwdetchar", detector=self.instrument
        )
        summary = report.Tab("Summary")

        logger.info("computing runtime overview")
        summary += self.create_runtime_overview()

        # generate a table summarizing the basic numbers for each nickname
        logger.info("computing basic statistics")
        summary += report.Header("Basic Statistics")
        summary += self.create_classifier_overview(
            nicknames,
            datasets,
            models,
            calibmaps,
            series=series,
        )

        # make plots
        # ROC curve
        logger.info("generating ROC plot")
        title = "Receiver Operating Characteristic Curves"
        path = plot_roc(
            self.config,
            self.start,
            self.end,
            nicknames,
            datasets,
            series=series,
            output_dir=reportdir,
        )
        logger.debug(f"saved plot to: {path}")
        summary += report.Header(title)
        summary += report.ImageBoard(title, [[path]])

        # series and calib stuff that depends on series
        calib_plots = []
        if series is not None:
            # timeseries
            logger.info("generating histograms")
            title = "Timeseries Histograms"
            path = plot_histogram(
                self.config,
                self.start,
                self.end,
                nicknames,
                series,
                datasets=datasets,
                output_dir=reportdir,
            )
            logger.debug(f"saved plot to: {path}")
            summary += report.Header(title)
            summary += report.ImageBoard(title, [[path]])

            # FIXME: add cumulative histograms of timeseries behavior

            # calibration coverage
            logger.info("generating calibration coverage plot")
            path = plot_calibration_coverage(
                self.config,
                self.start,
                self.end,
                nicknames,
                series,
                datasets,
                output_dir=reportdir,
            )
            calib_plots.append(path)
            logger.debug(f"saved plot to: {path}")

        # calibration distributions
        logger.info("generating calibration pdf/cdf plot")
        path = plot_calibration_pdf_cdf(
            self.config,
            self.start,
            self.end,
            nicknames,
            calibmaps,
            output_dir=reportdir,
        )
        calib_plots.append(path)
        logger.debug(f"saved plot to: {path}")

        summary += report.Header("Calibration")
        summary += report.ImageBoard("Calibration", [calib_plots])

        # add correlation plots
        if len(nicknames) > 1:
            logger.info("generating glitch/clean corner plots")
            gch_path, cln_path = plot_corner_glitch_clean(
                self.config,
                self.start,
                self.end,
                nicknames,
                datasets,
                output_dir=reportdir,
            )
            logger.debug(f"saved plots to: {gch_path}, {cln_path}")
            summary += report.Header("Comparisons")
            summary += report.ImageBoard("Comparisons", [gch_path, cln_path])

            # FIXME: add a corner for series as well?

        # write date into a footer
        summary += report.Footer("Report Metadata")
        summary += report.Description(
            "generated at %s"
            % time.strftime("%H:%M:%S %Z %a %d %b %Y", time.localtime())
        )
        summary += report.Description(
            "hostname : " + os.environ.get("HOSTNAME", "unknown")
        )
        summary += report.Description("username : " + os.environ.get("USER", "unknown"))
        thisreport += summary

        # generate single-classifier summary pages
        if not self.overview_only:
            comparisons = self.classifier_report(
                nicknames,
                datasets,
                models,
                calibmaps,
                series=series,
                output_dir=reportdir,
            )
            for comparison in comparisons:
                thisreport += comparison

        if not self.overview_only:
            target_times = find_target_features(
                self.start, self.end, self.config, self.segs, logger, n_events=10
            )
            events = self.event_report(
                nicknames,
                datasets,
                models,
                calibmaps,
                self.segs,
                series=series,
                logger=logger,
                output_dir=reportdir,
                times=target_times,
            )
            thisreport += events

        # write html page
        thisreport.save(reportdir, config_name="report_test")

    def classifier_report(
        self,
        nicknames,
        datasets,
        models,
        calibmaps,
        series=None,
        output_dir=".",
    ):
        """
        generate single-classifier summary information
        """
        comparisons = []
        for nickname in nicknames:
            logger.info("generating single-classifier HTML page for " + nickname)
            comparison = report.Tab(nickname)

            logger.info("computing runtime overview")
            comparison += self.create_runtime_overview()

            # generate a table summarizing the basic numbers for each nickname
            logger.info("computing basic statistics")
            comparison += report.Header("Basic Statistics")
            comparison += self.create_classifier_overview(
                [nickname],
                datasets,
                models,
                calibmaps,
                series=series,
            )

            # make plots
            # ROC curve
            logger.info("generating ROC plot")
            path = plot_roc(
                self.config,
                self.start,
                self.end,
                [nickname],
                datasets,
                series=series,
                output_dir=output_dir,
                annotate_auc=self.annotate_auc,
            )
            logger.debug(f"saved plot to: {path}")
            title = "Receiver Operating Characteristic (ROC) Curves"
            comparison += report.Header(title)
            comparison += report.ImageBoard(title, images=[[path]])

            # series and calib stuff that depends on series
            calib_plots = []
            if series is not None:
                # timeseries
                logger.info("generating histograms")
                title = "Timeseries Histograms"
                path = plot_histogram(
                    self.config,
                    self.start,
                    self.end,
                    [nickname],
                    series,
                    datasets=datasets,
                    output_dir=output_dir,
                )
                logger.debug(f"saved plot to: {path}")
                comparison += report.Header(title)
                comparison += report.ImageBoard(title, [[path]])

                # FIXME: add cumulative histograms of timeseries behavior

                # calibration coverage
                logger.info("generating calibration coverage plot")
                path = plot_calibration_coverage(
                    self.config,
                    self.start,
                    self.end,
                    [nickname],
                    series,
                    datasets,
                    output_dir=output_dir,
                )
                logger.debug(f"saved plot to: {path}")
                calib_plots.append(path)

            # calibration distributions
            logger.info("generating calibration pdf/cdf plot")
            path = plot_calibration_pdf_cdf(
                self.config,
                self.start,
                self.end,
                [nickname],
                calibmaps,
                output_dir=output_dir,
            )
            logger.debug(f"saved plot to: {path}")
            calib_plots.append(path)

            comparison += report.Header("Calibration")
            comparison += report.ImageBoard("Calibration", [calib_plots])

            # individual calibration distribs
            if self.single_calib_plots:
                logger.info(
                    "generating calibration plots for individual calibration maps"
                )
                if series is not None:
                    _, calibs_segdict = find_span_ids(series)
                else:
                    calibs_segdict = None

                for calibmap in calibmaps[nickname].values():
                    # add a section to calibration for each individual map
                    calib_name = f"{nickname}_{calibmap.hash}"

                    # calibration coverage for only the samples evaluated with
                    # this map
                    if series is not None:
                        dataset = datasets[nickname]
                        if calibs_segdict is not None:
                            # filter this to only contain times when which this
                            # calibmap was used
                            dataset = dataset.copy()
                            dataset.filter(
                                segs=(
                                    dataset.segs
                                    & calibs_segdict[nickname][calibmap.hash]
                                )
                            )

                        logger.info(f"generating coverage plot for ID: {calibmap.hash}")
                        coverage_path = plot_calibration_coverage(
                            self.config,
                            self.start,
                            self.end,
                            [calib_name],
                            {calib_name: series[nickname]},
                            {calib_name: dataset},
                            output_dir=output_dir,
                            calib_id=calibmap.hash,
                        )
                        logger.debug(f"saved plot to: {path}")
                    else:
                        coverage_path = None

                    # distribution showing only this map
                    logger.info(f"generating pdf/cdf plot for ID: {calibmap.hash}")
                    distrib_path = plot_calibration_pdf_cdf(
                        self.config,
                        self.start,
                        self.end,
                        [calib_name],
                        {calib_name: {calibmap.hash: calibmap}},
                        output_dir=output_dir,
                        calib_id=calibmap.hash,
                    )
                    logger.debug(f"saved plot to: {path}")

                    thisgrid = report.ImageGrid(
                        nickname + calibmap.hash, grid_size=2, visible=False
                    )
                    if coverage_path:
                        thisgrid += report.Image(path=coverage_path)
                    thisgrid += report.Image(path=distrib_path)
                    comparison += thisgrid

            # feature importance
            comparison += report.Header("Feature Importance")
            for i, (model_hash, model) in enumerate(models[nickname].items(), 1):
                logger.info(
                    "generating {}/{} feature importance plots".format(
                        i, len(models[nickname])
                    )
                )
                importance_paths, importance_cols_datas = plot_feature_importance(
                    self.config,
                    self.start,
                    self.end,
                    [nickname],
                    {nickname: {model_hash: model}},
                    datasets=datasets,
                    series=series,
                    t0=self.t0,
                    output_dir=output_dir,
                )

                for (model_name, path), (_, cols, data) in zip(
                    importance_paths, importance_cols_datas
                ):
                    # feature importance plots
                    if path is not None:
                        thisgrid = report.ImageGrid(
                            model_name, grid_size=2, visible=False
                        )
                        thisgrid += report.Image(path=path)
                        comparison += thisgrid

                    # veto configurations
                    if cols is not None:
                        comparison += report.FlatTable(
                            f"Veto Config: {model_name}", columns=cols, rows=data
                        )

            # add footer
            comparison += report.Footer("Report Metadata")
            comparison += report.Description(
                "generated at %s"
                % time.strftime("%H:%M:%S %Z %a %d %b %Y", time.localtime())
            )
            comparison += report.Description(
                "hostname : " + os.environ.get("HOSTNAME", "unknown")
            )
            comparison += report.Description(
                "username : " + os.environ.get("USER", "unknown")
            )

            comparisons.append(comparison)

        return comparisons

    def event_report(
        self,
        nicknames,
        datasets,
        models,
        calibmaps,
        segs,
        times,
        series=None,
        logger=None,
        output_dir=".",
    ):
        """
        generate single-event summary information
        """
        if logger is not None:
            logger.info("generating top significance events HTML page")
        events = report.Tab("Events")

        title = "Timeseries"
        events += report.Header(title)

        if series is not None:
            for trigger_time in times:
                window_start = trigger_time - 2.0
                window_end = trigger_time + 2.0
                path = plot_timeseries(
                    self.config,
                    window_start,
                    window_end,
                    nicknames,
                    series,
                    t0=trigger_time,
                    datasets=datasets,
                    segments=self.segs,
                    output_dir=output_dir,
                )
                title = f"{trigger_time} Timeseries"
                events += report.ImageBoard(title, [[path]])

            # add footer
            events += report.Footer("Report Metadata")
            events += report.Description(
                "generated at %s"
                % time.strftime("%H:%M:%S %Z %a %d %b %Y", time.localtime())
            )
            events += report.Description(
                "hostname : " + os.environ.get("HOSTNAME", "unknown")
            )
            events += report.Description(
                "username : " + os.environ.get("USER", "unknown")
            )

        return events


def find_timeseries(config, start, end, nickname):
    """Find all timeseries within start/end corresponding to a single nickname."""
    reporter_factory = factories.ReporterFactory()
    timeseriesdir = names.tag2timeseriesdir(config.tag, rootdir=config.rootdir)
    timeseriesreporter = reporter_factory(
        timeseriesdir, start, end, **config.timeseries["reporting"]
    )

    all_series = []
    for _, _, found_series in timeseriesreporter.glob(
        f"{config.instrument}-{nickname}", start, end
    ):
        for series in found_series:
            series.crop(start, end)
            if len(series):  # some remaining time
                all_series.append(series)

    return combine_series(all_series)


def find_datasets(config, start, end, nickname, segments=None):
    """Find all datasets within start/end corresponding to a single nickname."""
    reporter_factory = factories.ReporterFactory()
    evaluatedir = names.tag2evaluatedir(config.tag, rootdir=config.rootdir)
    evaluatereporter = reporter_factory(
        evaluatedir, start, end, **config.evaluate["reporting"]
    )

    datasets = features.Dataset(start=start, end=end, segs=segments)
    segs = segmentlist([segment(start, end)])

    for _, _, dataset in evaluatereporter.glob(nickname, start, end, run="*"):
        # trim to requested start/end times
        dataset.filter(segs=(dataset.segs & segs))
        if len(dataset) > 0:
            datasets += dataset

    # attach a dataloader to enable loading features as requested
    # (feature importance plots, etc)
    # FIXME: this is done in a roundabout way but this should be
    # refactored to not use private variables
    dataloader_factory = factories.DataLoaderFactory()
    dataloader = dataloader_factory(start, end, segs=segments, **config.features)
    datasets._dataloader = dataloader

    return datasets


def find_models(config, start, end, nickname, hashes=None):
    """Find all classifier models within start/end corresponding to a single
    nickname."""
    reporter_factory = factories.ReporterFactory()
    traindir = names.tag2traindir(config.tag, rootdir=config.rootdir)
    trainreporter = reporter_factory(traindir, start, end, **config.train["reporting"])

    models = {}
    for _, _, model in trainreporter.glob(nickname, start, end, run="*"):
        models[model.hash] = model

    if hashes is not None:
        models = filter_data_by_hash(models, hashes)
    return models


def find_calibration_maps(config, start, end, nickname, hashes=None):
    """Find all calibration maps within start/end corresponding to a single
    nickname."""
    reporter_factory = factories.ReporterFactory()
    calibratedir = names.tag2calibratedir(config.tag, rootdir=config.rootdir)
    calibratereporter = reporter_factory(
        calibratedir, start, end, **config.calibrate["reporting"]
    )

    calib_maps = {}
    for _, _, calib in calibratereporter.glob(nickname, start, end, run="*"):
        calib_maps[calib.hash] = calib

    if hashes is not None:
        calib_maps = filter_data_by_hash(calib_maps, hashes)
    return calib_maps


def find_span_ids(all_series):
    """Find the span for all models and calibration maps used within timeseries."""
    models_segdict = {
        nickname: defaultdict(segmentlist) for nickname in all_series.keys()
    }
    calibs_segdict = {
        nickname: defaultdict(segmentlist) for nickname in all_series.keys()
    }
    for nickname, series in all_series.items():
        for s in series:
            seg = segment(s.start, s.end)
            models_segdict[nickname][s.model_id].append(seg)
            calibs_segdict[nickname][s.calibration_id].append(seg)
        for model in models_segdict[nickname].keys():
            models_segdict[nickname][model].coalesce()
        for calib in calibs_segdict[nickname].keys():
            calibs_segdict[nickname][calib].coalesce()

    return models_segdict, calibs_segdict


def find_target_features(start, end, config, segments, logger, n_events):

    # grab data
    dataloader_factory = factories.DataLoaderFactory()
    dataloader = dataloader_factory(start, end, segs=segments, **config.features)

    target_channel = config.samples["target_channel"]
    target_bounds = configparser.config2bounds(config.samples["target_bounds"])

    data_table = dataloader.query(channels=target_channel, bounds=target_bounds)
    data_table_ftrs = data_table.features[target_channel]
    snr_idxs = data_table_ftrs.argsort(keys=config.features["significance"])

    data_table_ftrs_srt = data_table_ftrs[snr_idxs]
    top_snr_triggers = data_table_ftrs_srt[-n_events:]
    return [row[0] for row in top_snr_triggers]


def filter_data_by_hash(data, hashes):
    """Filter data products to only the identifiers (hashes) specified."""
    hashes = set(hashes)
    return {hash_: d for hash_, d in data.items() if hash_ in hashes}


def plot_path(name, instrument, tag, output_dir=".", suffix="png"):
    """Generate a standardized plot filename/path."""
    return os.path.join(
        output_dir, f"{name}-{instrument}-{__process_name__}-{tag}.{suffix}"
    )


def plot_histogram(
    config, start, end, nicknames, series, datasets=None, t0=None, output_dir="."
):
    """Create a histogram of the timeseries data and save it to disk"""
    if t0 is None:
        t0 = start

    fig = plots.histogram(nicknames, series, start, end, t0)
    fig.suptitle(
        f"{config.instrument} Timeseries Histogram\nwithin [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )

    path = plot_path(
        f"histogram-{'-'.join(nicknames)}-",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(path, fig)

    return path


def plot_timeseries(
    config,
    start,
    end,
    nicknames,
    series,
    datasets=None,
    t0=None,
    segments=None,
    output_dir=".",
):
    """Create a timeseries plot and saves it to disk."""
    if config.report.get("annotate_gch"):
        if not datasets:
            raise ValueError("datasets required when annotating glitches")
        gch_gps = []
        for dataset in datasets.values():
            gch_gps += list(dataset.vectors2classes()[0].times)
        gch_gps = sorted(set(gch_gps))
    else:
        gch_gps = None

    if t0 is None:
        t0 = start

    fig = plots.timeseries(
        nicknames, series, start, end, t0, gch_gps=gch_gps, segs=segments
    )
    fig.suptitle(
        f"{config.instrument} Timeseries\nwithin [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    path = plot_path(
        f"timeseries-{'-'.join(nicknames)}-%s-%s" % (int(start), int(end)),
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(path, fig)
    return path


def plot_corner_glitch_clean(config, start, end, nicknames, datasets, output_dir="."):
    """Create a glitch/clean corner plot and saves it to disk."""
    # divide up classes
    gch = dict()
    cln = dict()
    for nickname in nicknames:
        gch[nickname], cln[nickname] = datasets[nickname].vectors2classes()

    # plot glitch correlations
    fig = plots.dataset_corner(nicknames, gch)
    fig.suptitle(
        f"{config.instrument} Glitch Correlations\nwithin [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    gch_path = plot_path(
        f"glitch_corner-{'-'.join(nicknames)}",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(gch_path, fig)

    # plot clean correlations
    fig = plots.dataset_corner(nicknames, cln)
    fig.suptitle(
        f"{config.instrument} Clean Correlations\nwithin [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    cln_path = plot_path(
        f"clean_corner-{'-'.join(nicknames)}",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(cln_path, fig)

    return gch_path, cln_path


def plot_calibration_coverage(
    config, start, end, nicknames, series, datasets, output_dir=".", calib_id=None
):
    """Create a calibration coverage plot and saves it to disk."""
    fig = plots.calibration_coverage(
        nicknames, series, datasets, legend=config.report.get("legend", False)
    )
    fig.suptitle(
        f"{config.instrument} Calibration Coverage\nwithin [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    path = plot_path(
        f"calibration_coverage-{'-'.join(nicknames)}",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(path, fig)
    return path


def plot_calibration_pdf_cdf(
    config, start, end, nicknames, calibmaps, output_dir=".", calib_id=None
):
    """Create a PDF/CDF plot for glitch/clean distributions and saves it to
    disk."""
    fig = plots.calibration_distribs(nicknames, calibmaps)
    fig.suptitle(
        f"{config.instrument} Calibration Distributions\n"
        f"within [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    path = plot_path(
        f"calibration_distributions-{'-'.join(nicknames)}",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(path, fig)
    return path


def plot_roc(
    config,
    start,
    end,
    nicknames,
    datasets,
    series=None,
    output_dir=".",
    annotate_auc=False,
):
    """Create a ROC curve plot and saves it to disk."""
    fig = plots.roc(nicknames, series, datasets, annotate_auc=annotate_auc)
    fig.suptitle(
        f"{config.instrument} Receiver Operating Characteristics\n"
        f"within [{start:.3f}, {end:.3f})",
        fontsize=plots.DEFAULT_FONTSIZE,
    )
    path = plot_path(
        f"roc-{'-'.join(nicknames)}",
        config.instrument,
        config.tag,
        output_dir=output_dir,
    )
    plots.save(path, fig)
    return path


def plot_feature_importance(
    config,
    start,
    end,
    nicknames,
    models,
    datasets=None,
    series=None,
    t0=None,
    output_dir=".",
):
    """Create a feature importance plot and saves it to disk."""
    if series:
        segdict, _ = find_span_ids(series)
    else:
        segdict = None

    # instantiate classifiers to pick up some standard stuff...
    classifier_factory = factories.ClassifierFactory()
    klassifiers = {
        nickname: classifier_factory(
            nickname, rootdir=output_dir, **config.classifier_map[nickname]
        )
        for nickname in nicknames
    }

    figs, tabs = plots.featureimportance(
        nicknames, models, klassifiers, datasets, start, end, t0=t0, segdict=segdict
    )
    paths = []
    cols_datas = []
    for nickname in nicknames:
        for key, model in models[nickname].items():
            model_name = f"{nickname}_{model.hash}"

            fig = figs[nickname][key]
            tab = tabs[nickname][key]
            if fig is not None:
                path = plot_path(
                    f"featureimportance-{nickname}-{model.hash}",
                    config.instrument,
                    config.tag,
                    output_dir=output_dir,
                )
                plots.save(path, fig)
                paths.append((model_name, path))
            else:
                paths.append((model_name, None))

            if tab is not None:
                cols, data = tab
                cols_datas.append((model_name, cols, data))
            else:
                cols_datas.append((model_name, None, None))

    return paths, cols_datas
